"""DO NOT EDIT.

This file was autogenerated. Do not edit it by hand,
since your modifications would be overwritten.
"""

from keras.src.optimizers import LegacyOptimizerWarning as Adagrad
from keras.src.optimizers import LegacyOptimizerWarning as Adam
from keras.src.optimizers import LegacyOptimizerWarning as Ftrl
from keras.src.optimizers import LegacyOptimizerWarning as Optimizer
from keras.src.optimizers import LegacyOptimizerWarning as RMSprop
from keras.src.optimizers import LegacyOptimizerWarning as SGD
